/* $Id: conf.c,v 1.90 2004/05/24 12:29:35 rav Exp $	*/

/* Intro {{{
 * ----------------------------------------------------------------
 *  DConnect Daemon
 *
 *	  #(@) Copyright (c) 2002, DConnect development team
 *	  #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 *  This  program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You  should  have  received a copy  of  the GNU General Public
 *  License  along  with  this  program;  if  not,  write  to  the
 *  Free  Software  Foundation,  Inc.,  675  Mass  Ave,  Cambridge,
 *  MA 02139, USA.
 *
 * ----------------------------------------------------------------
 * }}} */

/* Partially borrowed from GNU Wget
 * Modified by rawsock */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "match.h"
#include "misc.h"
#include "net.h"
#include "penalties.h"
/* }}} */

/* commands recognized by config {{{ */
const char *commands[]=
{
	"listen_port_main",	// 0
	"listen_port_cons",	// 1
	"hubname",			// 2
	"user_limit",		// 3
	"allow_broken_key",	// 4
	"ping_timeout",		// 5
	"register_timeout",	// 6
	"conf_nallow",		// 7
	"min_share",		// 8
	"log_dir",			// 9
	"log_main",			// 10
	"user",				// 11
	"group",			// 12
	"allow_chat",		// 13
	"conf_welcome",		// 14
	"conf_motd",		// 15
	"conf_cusers",		// 16
	"conf_callow",		// 17
	"conf_banned",		// 18
	"conf_penalties",	// 19
	"log_level",		// 20
	"debug_level",		// 21
	"log_net",			// 22
	"log_proto",		// 23
	"log_wrng",			// 24
	"log_chat",			// 25
	"log_con",			// 26
	"min_slots",		// 27
	"allow_passive",	// 28
	"std_penalty_duration", //29
	"listen_interface", //30
	"admin_contact",  //31
	""
}; /* }}} */

extern struct config conf;

extern int NPENAL; //
extern Tpenalty *penalties[MAXPENALTIES];

/* strdupdelim() - copy the string formed by two pointers to a new location {{{ */
char *strdupdelim(const char *beg,const char *end) {
	char *res=(char *)my_malloc(end-beg+1);
	
	memcpy(res,beg,end-beg);
	res[end-beg]=0;
	return res;
} /* }}} */

/* Syntax: <sp>* command <sp>* = <sp>* value <newline>
 * Return values:
 *	1	- success
 *	0	- failure
 *	-1	- empty */ 
/* parse_line() {{{ */
int parse_line(const char *line,char **com,char **val) 
{
	const char *p=line;
	const char *orig_comptr,*end;
	char *new_comptr;
	
	while(*p&&isspace(*p)) ++p;
	
	if (!*p||*p=='#') return -1;
	
	for (orig_comptr=p;isalpha(*p)||*p=='_'||*p=='-';p++);
	
	if (!isspace(*p)&&(*p!='=')) return 0;
	
	*com=(char *)my_malloc(p-orig_comptr+1);
	
	for (new_comptr=*com;orig_comptr<p;orig_comptr++) *new_comptr++ = *orig_comptr;
	*new_comptr=0;

	for (;isspace(*p);p++);
	
	if (*p!='=') 
	{ 
		my_free (*com); 
		return 0;
	}
	
	for (++p;isspace(*p);p++);
	
	end=line+strlen(line)-1;
	
	while(end>p&&isspace(*end)) --end;
	
	*val=strdupdelim(p,end+1);

	return 1;
} /* }}} */

/* read_whole_line() - read a line from FP and return the pointer to allocated storage. {{{ */
char *read_whole_line(FILE *fp) 
{
	int length=0;
	int bufsize=82;
	char *line=(char *)my_malloc(bufsize);
	
	while(fgets(line+length,bufsize-length,fp))	
	{
		length+=strlen(line+length);
		
		if (length==0) continue;
		if (line[length-1]=='\n') break;
		
		bufsize<<=1;
		
		line=(char *)my_realloc(line,bufsize);
	}
		
	if (length==0||ferror(fp)) 
	{
		my_free (line);
		return NULL;
	}
	
	if (length+1<bufsize) line=(char *)my_realloc(line,length+1);
	return line;
} /* }}} */

/* parse_config() - parse config file {{{ */
char parse_config(struct config *conf) 
{
	char *line,tmp[PATH_MAX];
	int ln;
	FILE *fp;
	struct passwd *user_data;
	struct group *group_data;
	char *com=NULL,*val=NULL;
	int status;

	/* opening config {{{ */
	/* we look for config in location taken from configure or from command-line option */
	if (!realpath(conf->conf_main,tmp)) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"Can't find config file (%s): %s.",conf->conf_main, strerror(errno));
		return 1;
	}
	
	my_free(conf->conf_main);
	conf->conf_main=strdup(tmp);
	fp=fopen(conf->conf_main,"rb");
	
	if (!fp) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"Can't open config file (%s): %s.",conf->conf_main, strerror(errno));
		return 1;
	}
	/* }}} */

	ln=1;
	while((line=read_whole_line(fp))) 
	{
		status=parse_line(line,&com,&val);
		my_free(line);
		
		if (status==1) 
		{
			status=0;
		
			while(*commands[status]) 
			{
				if (!strcasecmp(com,commands[status])) break;
				else status++;
			}
			
			/* Options {{{ */
			switch(status)
			{
				/* 0. listen_port_main {{{ */
				case 0:
						if ((status=atoi(val))<=0)
						{
							debug(DEBUG_STD,"Error in %s at line %d.",conf->conf_main,ln);
							goto leave;
						}
			
						conf->listen_port_main=status;
						conf->listen_port_sr=status+1;
						
						break; /* }}} */
				
				/* 1. listen_port_cons {{{ */
				case 1: 
						conf->listen_port_cons=atoi(val);
						break; /* }}} */
				
				/* 2. hubname {{{ */
				case 2: 
						my_free(conf->hubname);
						conf->hubname=strdup(val);
						break; /* }}} */
				
				/* 3. user_limit {{{ */
				case 3: 
						conf->userlimit=atoi(val);
						break; /* }}} */

				/* 4. allow_broken_key {{{ */
				case 4: 
						conf->allow_broken_key=atoi(val);
						break; /* }}} */
				
				/* 5. ping_timeout {{{ */
				case 5: 
						conf->ping_timeout=atoi(val);
						break; /* }}} */
				
				/* 6. register_timeout {{{ */
				case 6: 
						conf->register_timeout=atoi(val);
						break; /* }}} */
				
				/* 7. allowed nicks file {{{ */
				case 7: 
						my_free(conf->conf_nallow);
						conf->conf_nallow=strdup(val);
						break; /* }}} */
				
				/* 8. min_share {{{ */
				case 8: 
						conf->minshare=atoi(val);
						break; /* }}} */
				
				/* 9. log files dir {{{ */
				case 9: 
						my_free(conf->log_dir);
						conf->log_dir=strdup(val);
						break; /* }}} */
				
				/* 10. main log file name {{{ */
				case 10: 
						my_free(conf->log_main);
						conf->log_main=strdup(val);
						break; /* }}} */
				
				/* 11. UID to run {{{ */
				case 11: 
						user_data=getpwnam(val);
				
						if(!user_data) 
						{
							debug(DEBUG_STD,"Error in %s at line %d: user `%s' not found.",conf->conf_main,ln,val);
							goto leave;
						}
					
						conf->uid=user_data->pw_uid;
						break; /* }}} */
				
				/* 12. GID to run {{{ */
				case 12: 
						group_data=getgrnam(val);
				
						if(!group_data) 
						{
							debug(DEBUG_STD,"Error in %s at line %d: group `%s' not found.",conf->conf_main,ln,val);
							goto leave;
						}
				
						conf->gid=group_data->gr_gid;
						break; /* }}} */
				
				/* 13. allow chat {{{ */
				case 13: 
						conf->allow_chat=atoi(val);
						break; /* }}} */
				
						/* 14. welcome message file {{{ */
				case 14: 
						my_free(conf->conf_welcome);
						conf->conf_welcome=strdup(val);
						break; /* }}} */
				
				/* 15. message of the day {{{ */
				case 15: 
						my_free(conf->conf_motd);
						conf->conf_motd=strdup(val);
						break; /* }}} */
				
				/* 16. dcd.users {{{ */
				case 16: 
						my_free(conf->conf_cusers);
						conf->conf_cusers=strdup(val);
						break; /* }}} */
				
				/* 17. console.allow {{{ */
				case 17: 
						my_free(conf->conf_callow);
						conf->conf_callow=strdup(val);
						break; /* }}} */
				
				/* 18. bans file {{{ */
				case 18: 
						my_free(conf->conf_banned);
						conf->conf_banned=strdup(val);
						break; /* }}} */

				/* 18. penalties file {{{ */
				case 19: 
						my_free(conf->conf_penalties);
						conf->conf_penalties=strdup(val);
						break; /* }}} */
				
				/* 20. log level {{{ */
				case 20: 
						conf->log_level=atoi(val);
						break; /* }}} */
				
				/* 21. debug level {{{ */
				case 21: 
						if ((status=atoi(val))<=0) 
						{
							debug(DEBUG_STD,"Error in %s at line %d.",conf->conf_main,ln);
							goto leave;
						}
						
						if (status&DEBUG_PROTO) status=status-DEBUG_PROTO;
						
						conf->debug_level=status;
						
						break; /* }}} */
				
				/* 22. net log file name {{{ */
				case 22: 
						my_free(conf->log_net);
						conf->log_net=strdup(val);
						break; /* }}} */
				
				/* 23. proto log file name {{{ */
				case 23: 
						my_free(conf->log_proto);
						conf->log_proto=strdup(val);
						break; /* }}} */
				
				/* 24. wrng log file name {{{ */
				case 24: 
						my_free(conf->log_wrng);
						conf->log_wrng=strdup(val);
						break; /* }}} */
				
				/* 25. chat log file name {{{ */
				case 25: 
						my_free(conf->log_chat);
						conf->log_chat=strdup(val);
						break; /* }}} */
				
				/* 26. console log file name {{{ */
				case 26: 
						my_free(conf->log_con);
						conf->log_con=strdup(val);
						break; /* }}} */

				/* 27. Minimal Amount of Slots {{{ */
				case 27: 
						conf->minslots=atoi(val);
						break; /* }}} */

				/* 28. if passive mode is allowed {{{ */
				case 28: 
						conf->allow_passive=atoi(val);
						break; /* }}} */

				case 29: 
						conf->std_penalty_duration=atol(val);
						break; /* }}} */

				/* 30. specifying listening interface*/
				case 30: 
						my_free(conf->listen_interface);
						conf->listen_interface=strdup(val);
						break; /* }}} */

				/* 31. specifying contact to admin which is shown when user is not allowed to connect to hub */
				case 31: 
						my_free(conf->admin_contact);
						conf->admin_contact=strdup(val);
						break; /* }}} */


				/* action when everything fails: {{{ */
				default: 
						debug(DEBUG_STD,"Unknown keyword in %s at line %d: %s.",conf->conf_main,ln,com);
						goto leave; /* }}} */
			} /* }}} */
		
			my_free(com);
			my_free(val);

		}
		else 
			if (status==0) 
			{
				debug(DEBUG_STD,"Error in %s at line %d.",conf->conf_main,ln);
				fclose(fp);
				return 1;
			}

		++ln;
	}

	fclose(fp);
	return 0;


leave:
	my_free(com);
	my_free(val);
	fclose(fp);
	return 1;
} /* }}} */

/* playfile() - parse text file and play to public chat window {{{ */
void playfile(struct userrec *usr,char *msgfile) 
{
	char *line,*p;
	FILE *fp;
	char *__strtok_temp__;

	seteuid(0);
	setegid(0);
	fp=fopen(msgfile,"rb");
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't open message file (%s): %s.",msgfile,strerror(errno));
		return;
	}

	while((line=read_whole_line(fp))) 
	{
		p=line;
		while(*p&&isspace(*p)) ++p;
	
		if (!*p||*p=='#') 
		{
			my_free(line); 
			continue; 
		}
	
		strtok_r(p,"\r\n",&__strtok_temp__);
		
		pubmsg(usr,p);
		my_free(line);
	}

	fclose(fp);
} /* }}} */

/* validlogin() - verify login and password {{{ */
char validlogin(struct userrec *usr,char *login,char *pass)
{
	const char *sep=":\r\n";
	char *line,*p,*tmp;
	FILE *fp;

	seteuid(0);
	setegid(0);
	fp=fopen(conf.conf_cusers,"rb");
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't open users file (%s): %s.",conf.conf_cusers,strerror(errno));
		return 0;
	}

	while((line=read_whole_line(fp))) 
	{
		p=line;
	
		while(*p&&isspace(*p)) ++p;
	
		if (!*p||*p=='#')
		{ 
			my_free(line);
			continue;
		}
	
		tmp=strpbrk(p,sep);
		if (!tmp) goto leave;
		*(tmp++)=0;

		if (!strcasecmp(p,login))
		{
			p=tmp;
			tmp=strpbrk(p,sep);
			if (!tmp) goto leave;
			*(tmp++)=0;
		 
			if (pass!=NULL)
			{
				if(strcmp(p,pass)) return 0;//console user with invalid password
		
				debug(DEBUG_NET,"'%s' from %s is Console User",login,usr->ip);
		
				p=tmp;
				tmp=strpbrk(p,sep);
				if (!tmp) goto leave;
				*(tmp++)=0;
			
				my_free(usr->email);
				usr->email=strdup(p);

				p=tmp;
				tmp=strpbrk(p,sep);
				if (!tmp) goto leave;
				*(tmp++)=0;
		
				my_free(usr->desc);
				usr->desc=strdup(p);

				p=tmp;
				tmp=strpbrk(p,sep);
				if (tmp) *tmp=0;
				usr->perm=strdup(p);

				my_free(line);
				fclose(fp);
				return 1;
			} 
			else 
				{
					debug(DEBUG_NET,"'%s' is a DC User with restricted nick",login);

					usr->password=strdup(p); //copy password

					p=tmp;
					tmp=strpbrk(p,sep);
					if (!tmp) goto leave;
					*(tmp++)=0;
					p=tmp;
					tmp=strpbrk(p,sep);
					if (!tmp) goto leave;
					*(tmp++)=0;
					p=tmp;
					tmp=strpbrk(p,sep);
					if (tmp) *tmp=0;

					if (p) usr->perm=strdup(p); //copy permissions
					else usr->perm=strdup("n");

					my_free(line);
 					fclose(fp);
					return 1; 
				}
		}
	
		my_free(line);
	}

leave:
	
		debug(DEBUG_NET,"'%s' is a DC user",login);
	
		usr->perm=strdup("n");
		my_free(line);
	
		fclose(fp);
		return 0;
} /* }}} */


/* validhost(usr, access_file, deeper){{{
.description:
	verifies host access
   
.args
	struct userrec *usr	userrec
	char *access_file	file name 
	int deeper		if !=0 then use include: lines in the acces_file
  
.return values:
	0:	access file does not exists, 
	host is not listed in the file
	
	1:	host is listed in the file
*/

char validhost(struct userrec *usr,char *access_file, int deeper) 
{
	const char *sep=":\r\n";

	char *line,*p;
	char *fpath;
	FILE *fp;
	char *__strtok_temp__;

	seteuid(0);
	setegid(0);

	/* open the access_file */
	fp=fopen(access_file,"rb");
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't open allow/deny file (%s): %s.",access_file,strerror(errno));
		return 0;
	}

	/* parse the line */
	while((line=read_whole_line(fp))) /* reads the lines */
	{
		p=line; /* p contains the line */

		while(*p&&isspace(*p)) ++p; 
 		if (!*p||*p=='#') { my_free(line); continue; } /* skip empty and # lines; goes to the next line*/

		p=strtok_r(p,sep,&__strtok_temp__); /* read the first token of the line*/
	
		if (!p) continue;

		/* check for include: */
		if (deeper && !strcasecmp(p,"include"))	/* deeper!=0 */
		{
			/* parse the include:>filename< */
			p=strtok_r(NULL,sep,&__strtok_temp__);  /* get the included file name */		   
		
			if (p)
			{
				fpath=strdup(p); 
				make_path(CONFIGDIR,&(fpath));
			
				if (strcasecmp(fpath,access_file))/* file name should != NULL neither access_file */
				{
					if (validhost(usr,fpath,0)) 
					{
						my_free(fpath);
						goto leave;
					}//valid
				}else debug(DEBUG_STD,"Including the same file is not allowed"); 

				my_free(fpath);

			} else debug(DEBUG_STD,"Syntax error: include:name_of_file"); 

		} else/* if ip mathes one of the listed in file ips */
			if (match(p,usr->ip)) 
			{
				p=strtok_r(NULL,sep,&__strtok_temp__);	   /*getting reason of ban */
				if(!usr->cons)
				{
					if (p) pubmsg(usr,p);  /* if the reason is specified then display it, else display "Your..."*/
					else pubmsg(usr,"Your host is banned.");
				}
				goto leave;
			}

		my_free(line);

	}//while

	my_free(line);
	fclose(fp);
	return 0;

leave:

	my_free(line);
	fclose(fp);
	return 1;

} /* }}} */


/* nick_allowed() - verify nick access {{{ */
char nick_allowed(struct userrec *usr,char *nick) 
{
	const char *sep=" \t\r\n";
	char *line=NULL,*p=NULL;
	FILE *fp;
	char *__strtok_temp__;

	seteuid(0);
	setegid(0);
	fp=fopen(conf.conf_nallow,"rb");
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't open nicks allow file (%s): %s.",conf.conf_nallow,strerror(errno));
		return 1; // allow by default
	}

	while((line=read_whole_line(fp))) 
	{
		p=line;
	
		while(*p&&isspace(*p)) ++p;

		if (!*p||*p=='#') 
		{ 
			my_free(line); 
			continue; 
		}
	
		p=strtok_r(p,sep,&__strtok_temp__);
		if (!p) goto leave;
	
		if (match(p,nick)) 
		{
			my_free(line);
			fclose(fp);
			return 1;
		}
	
		my_free(line);
	}

leave:
	my_free(line);
	fclose(fp);
	if (p) pubmsg(usr,"Your nick doesn't match any of the allowed patterns.");
	return 0;
} /* }}} */

/* penalties_read() - read penalties list from file to memory {{{ */
void penalties_read() 
{
	const char *sep=":\r\n";
	char *line=NULL,*p=NULL,*s_ip=NULL,*s_penalties, *s_start_date,*s_op,*s_nick,*s_duration,*s_reason;
	FILE *fp;
	int i=-1;
	long i_duration; 
	unsigned long i_start_date;
	int i_penalties;
	char *__strtok_temp__;

	debug(DEBUG_PROTO,"penalties_read() enter");
	seteuid(0);
	setegid(0);
	fp=fopen(conf.conf_penalties,"rb");
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't open file (%s): %s.",conf.conf_penalties,strerror(errno));
		goto leave;
	}

	while((line=read_whole_line(fp))) 
	{
		p=line;
		i++;
	
		while(*p&&isspace(*p)) ++p;

		if (!*p||*p=='#') 
		{ 
			my_free(line); 
			continue; 
		}
	
		s_ip=strtok_r(line,sep,&__strtok_temp__);
		s_nick=strtok_r(NULL,sep,&__strtok_temp__);
		s_op=strtok_r(NULL,sep,&__strtok_temp__);
		s_penalties=strtok_r(NULL,sep,&__strtok_temp__);
		s_start_date=strtok_r(NULL,sep,&__strtok_temp__);
    	s_duration=strtok_r(NULL,sep,&__strtok_temp__);
		s_reason=strtok_r(NULL,"\r\n",&__strtok_temp__);

		if (!s_ip || !s_penalties || !s_duration || !s_reason || 
			!sscanf(s_start_date,"%lu",&i_start_date) ||
			!sscanf(s_duration,"%ld",&i_duration) || 
			!sscanf(s_penalties,"%d",&i_penalties))
		{
			debug(DEBUG_WRNG,"%s(%d): Syntax Error",conf.conf_penalties,i);
			my_free(line); 
			continue; 
		} 

		user_set_penalty(s_ip,s_nick,s_op,i_penalties,i_start_date,i_duration,s_reason,1);
	
		my_free(line);	
	}

leave:
	debug(DEBUG_PROTO,"penalties_read() leave");
	my_free(line);
	fclose(fp);

} /* }}} */

/* penalties_write() - write penalties list from memory to file {{{ */
void penalties_write() 
{
	FILE *fp;
	long i,duration;

	debug(DEBUG_PROTO,"penalties_write() enter");
	seteuid(0);
	setegid(0);
	fp=fopen(conf.conf_penalties,"w+");//clearing the file while opening
	seteuid(conf.uid);
	setegid(conf.gid);

	if (!fp) 
	{
		debug(DEBUG_STD,"Can't write penalties file (%s): %s.",conf.conf_penalties,strerror(errno));
		goto leave;
	}

	for(i=0;i<NPENAL;i++) 
	{
		if(penalties[i])
		{
			duration=user_exp_penalty(penalties[i]);

	//		debug(DEBUG_PROTO,"%s %d %ld %s",penalties[i]->ip,duration,penalties[i]->penalty,penalties[i]->reason);

			if (duration>0) fprintf(fp,"%s:%s:%s:%d:%lu:%ld:%s\n",penalties[i]->ip,
																  penalties[i]->nick,
																  penalties[i]->op,
																  penalties[i]->penalty,
																  penalties[i]->start_date,
																  duration,
																  penalties[i]->reason);
			else delpenalty(i,1);
		}
	}
leave:
	fclose(fp);
	debug(DEBUG_PROTO,"penalties_write() leave");
} /* }}} */




/* VIM Settings {{{
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * soft-stop-width: 4
 * c indent on
 * End:
 * vim600: sw=4 ts=4 sts=4 cindent fdm=marker
 * vim<600: sw=4 ts=4
 * }}} */
