/* $Id: main.c,v 1.274 2004/05/24 12:29:35 rav Exp $ */

/* Intro {{{
 * ---------------------------------------------------------------- 
 *  DConnect Daemon
 * 
 *	  #(@) Copyright (c) 2002, DConnect development team
 *	  #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 *  This  program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You  should  have  received a copy  of  the GNU General Public
 *  License  along  with  this  program;  if  not,  write  to  the
 *  Free  Software  Foundation,  Inc.,  675  Mass  Ave,  Cambridge,
 *  MA 02139, USA. 
 *
 * ---------------------------------------------------------------- 
 * }}} */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "conf.h"
#include "misc.h"
#include "net.h"
#include "match.h"
#include "penalties.h"
#include "console.h"
#include "cmd.user.h"

/* }}} */

/* reserved_nicks[] - nicks that can't be owned by normal users {{{ */
const char *reserved_nicks[]=
{
	"Hub",
	"Hub-Security",
	"Vandel\\Debug",
	"Client",
	"" /* indicates end of list */
}; /* }}} */


int NPENAL; /* penalties count */
int NUSERS; /* user count */

Tpenalty *penalties[MAXPENALTIES]; /* table of penalties */
struct userrec *user[MAXUSERS];	/* users table */
struct config conf;		/* configuration data */

pthread_t listen_th_main;	/* listen thread handle for main service */
pthread_t listen_th_cons;	/* listen thread handle for remote console */
pthread_t listen_th_sr;		/* listen thread handle for minslots checking */

pthread_mutex_t mutex;		/* provides synchronization between threads */
pthread_mutex_t mutex_inet_ntoa; // securing inet_ntoa

fd_set sockfdset;			/* descriptors */
int listen_main;			/* main listening socket */
int listen_cons;			/* console listening socket */
int listen_sr;				/* minslots checking socket */
char DAEMON;				/* set if running in background */
time_t starttime;			/* start time in sec */

/* logging */
FILE* file_net;
FILE* file_proto;
FILE* file_wrng;
FILE* file_con;
FILE* file_chat;

/* these are for tcpd */
int allow_severity;
int deny_severity;

int test_penalty_counter=0;

#ifdef HAVE___PROGNAME
extern char *__progname;
#else
char *__progname;
#endif


int user_set_state(struct userrec *usr, int state)
{
	if (usr->nick) debug(DEBUG_PROTO,"'%s' state changed from %d to %d.",usr->nick,usr->state,state);
	usr->state=state;
	return 0;
}

int user_tst_state(struct userrec *usr, int state)
{
//	if (usr->nick) debug(DEBUG_PROTO,"TesT '%s' state %d & %d.",usr->nick,usr->state,state);
	return ((usr->state)&state)>0;
}

/* adduser() - allocate and initialize new DC user {{{ */
void adduser(int sock,char *remote) 
{
	char lock[101], pk[17];
	struct userrec *tmp=NULL;
	struct sockaddr_in addr;
	int size_addr=sizeof(struct sockaddr_in);

	debug(DEBUG_PROTO,"adduser() enter");
	
	tmp=(struct userrec *)my_malloc(sizeof(struct userrec));
	bzero(tmp,sizeof(struct userrec));

	user_set_state(tmp,STATE_CONNECTED);
	tmp->ip=remote;
	tmp->sock=sock;
	tmp->slots=0;

	getsockname(tmp->sock,(struct sockaddr *)&addr,&size_addr);
	
	/* getting user's ip address */
	tmp->con_ip=my_inet_ntoa(addr.sin_addr);
	
	if (setsockopts(sock,0)) debug(DEBUG_WRNG,"setsockopts() failed.");

	/* we use select() anyway, but let's care for blocking send() */
	if (fcntl(sock,F_SETFL,O_NONBLOCK)<0) debug(DEBUG_WRNG,"fcntl() failed: %s.", strerror(errno));


	/* testing if user is listed in hosts.allow */
	if (denied(sock)) 
	{
		pubmsg(tmp,"Contact with %s",conf.admin_contact);
		disconnect(tmp,"You're not allowed to connect."); 
		goto leave; 
	}

	/* client complains that it couldn't find $Lock when msgs are sent before it */
	lockgen(lock,pk);

	sendf(tmp,"$Lock %s Pk=%s|",lock,pk);

	sendf(tmp,"$HubName %s|",conf.hubname);
	if (*(conf.conf_welcome)) playfile(tmp,conf.conf_welcome);
	

	/* host is banned permanently*/
	if (validhost(tmp,conf.conf_banned,1)) 
	{
		disconnect(tmp,NULL); 
		goto leave; 
	}
	
	if((NUSERS>=conf.userlimit)||(NUSERS+1>=MAXUSERS)) /* '+1' reserves 1 slot for admin console */
	{ 
		disconnect(tmp,"Sorry, this Hub is currently full!"); 
		goto leave; 
	}

	tmp->ping=time(NULL);
	tmp->flag=1;
	tmp->perm=NULL;
	tmp->password=NULL;

	pthread_mutex_lock( &mutex );
	user[NUSERS++]=tmp;
	FD_SET(sock,&sockfdset);
	pthread_mutex_unlock( &mutex );

leave:
	debug(DEBUG_PROTO,"adduser() leave");
} /* }}} */

/* deluser() - free normal or console user {{{ */
void deluser(struct userrec *usr,char *reason) 
{
	int id=usr2id(usr);
	
	if (!usr) return;

	pthread_mutex_lock(&mutex);
	user[id]=user[--NUSERS];
	user[NUSERS]=NULL;
	FD_CLR(usr->sock,&sockfdset);

	if (user_tst_state(usr,STATE_SHOW_QUIT)) 
	{
		sendf(NULL,"$Quit %s|",usr->nick);
		if (conf.flags & SHOW_JQ) sendf(NULL,"## %s has quit|",usr->nick);
	}
	disconnect(usr,reason);	
	pthread_mutex_unlock(&mutex);
} /* }}} */

/* listen_thread_maiN() - listen for new connections to main service {{{ */
void *listen_thread_main(void *args) 
{
	int new_sock,remote_size;
	struct sockaddr_in local,remote;

	char *ip;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	/* setup and bind the listening socket */
	if ((listen_main=socket(AF_INET,SOCK_STREAM,0))==-1) 
	{
		debug(DEBUG_SYS|DEBUG_STD,0,"Socket error - terminating: %s.", strerror(errno));
		exit(0);
	}
	
	if (setsockopts(listen_main,1)) debug(DEBUG_WRNG,0,"setsockopts() failed.");

	bzero(&local,sizeof(struct sockaddr_in));
	local.sin_family=AF_INET;
	local.sin_port=htons(conf.listen_port_main);
	inet_aton(conf.listen_interface,&local.sin_addr);

	seteuid(0);
	setegid(0);
	
	if (bind(listen_main,(struct sockaddr *)&local,sizeof(struct sockaddr))<0) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"Can't bind the main listening socket - terminating: %s.", strerror(errno));
		exit(0);
	}
		
	seteuid(conf.uid);
	setegid(conf.gid);
	
	if (listen(listen_main,-1)<0) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"listeN() failed - terminating: %s.", strerror(errno));
		exit(0);
	}

	/* accept connections */
	while(1)
	{
		remote_size=sizeof(struct sockaddr_in);
		
		if ((new_sock=accept(listen_main,(struct sockaddr *)&remote,&remote_size))<0)
		{
			if (errno!=EINTR) debug(DEBUG_WRNG,"accept() failed: %s.", strerror(errno));
			continue;
		}
		
		debug(DEBUG_NET,"Connection to DC from %s:%d.",ip=my_inet_ntoa(remote.sin_addr),ntohs(remote.sin_port));
		adduser(new_sock,ip);
	}
	
	debug(DEBUG_WRNG,"listen_thread_main: this should be never reached!");
	pthread_cancel(listen_th_main); /* never reached */
} /* }}} */

int penalty_welcome_banned(struct userrec *usr)
{
	int time_, days, hours,sec,min;

	/*sOP cannot be punished*/
	if (usr->perm && strchr(usr->perm,'s')) return 0;

	/* get the penalties for the user*/
	user_penalty(usr);

	/* host is banned temporaily by penalty */
	if (user_tst_penalty(usr,PENALTY_BANNED))
	{
		time_=user_exp_penalty(usr->penalty);

		days=time_/(long)(24*60*60);
		time_=time_%(24*60*60);
		
		hours=time_/(long)(60*60);
		time_=time_%(60*60);
		
		min=time_/(long)60;
		time_=time_%60;
		
		sec=time_;

		pubmsg(usr,"Yor host is banned. Contact with Administrator");
		pubmsg(usr,"Time left: %ld days and %ld:%ld:%ld",days,hours,min,sec);
		pubmsg(usr,"REASON: %s",usr->penalty->reason);
		user_set_state(usr,STATE_QUIT);
		return 1;
	}

	return 0;
}

void GetNickList(struct userrec *usr)
{
	char ops[(NICK_LEN+1)*MAXUSERS+1],nicks[(NICK_LEN+1)*MAXUSERS+1];
	int i;

	ops[0]='\000';
	nicks[0]='\000';		

	user_set_state(usr,STATE_REGISTERED);

	for(i=0;i<NUSERS;i++)
		if (user[i] && user_tst_state(user[i],STATE_REGISTERED)) 
		{
//			sendf(usr,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",user[i]->nick,user[i]->desc,user[i]->ctype,user[i]->flag,user[i]->email,user[i]->share);

			strcat(nicks,user[i]->nick);
			strcat(nicks,"$$");

			if (user[i] && (strchr(user[i]->perm,'s') || strchr(user[i]->perm,'a')))
			{
				strcat(ops,user[i]->nick);
				strcat(ops,"$$");
			}
		}
	//if user is not registered
	if (!user_tst_state(usr,STATE_REGISTERED)) 
	{
		strcat(nicks,usr->nick);
		strcat(nicks,"$$");

		if (strchr(usr->perm,'a')||strchr(usr->perm,'s'))
		{
			strcat(ops,usr->nick);
			strcat(ops,"$$");
		}
	}

	sendf(usr,"$NickList %s|",nicks);
	sendf(usr,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",usr->nick,usr->desc,usr->ctype,usr->flag,usr->email,usr->share);

	if (strchr(usr->perm,'s') || strchr(usr->perm,'a') ) sendf(NULL,"$OpList %s|",ops);
	else sendf(usr,"$OpList %s|",ops);

	sendf(NULL,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",usr->nick,usr->desc,usr->ctype,usr->flag,usr->email,usr->share);
}


void *listen_thread_SR(void *args) 
{
	int remote_size,n, slots_old;
	struct sockaddr_in local,remote;
	struct userrec *usr=NULL;
	char *ip=NULL, bufor[10001], *cmd=NULL, *nick=NULL, *s_slots=NULL, *__strtok_temp__;
	
	
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	/* setup and bind the listening socket */
	if ((listen_sr=socket(AF_INET,SOCK_DGRAM,0))==-1) 
	{
		debug(DEBUG_SYS|DEBUG_STD,0,"Socket error - terminating: %s.", strerror(errno));
		exit(0);
	}
	
	if (setsockopts(listen_sr,1)) debug(DEBUG_WRNG,0,"setsockopts() failed.");

	bzero(&local,sizeof(struct sockaddr_in));
	local.sin_family=AF_INET;
	local.sin_port=htons(conf.listen_port_sr);

	inet_aton(conf.listen_interface,&local.sin_addr);

	seteuid(0);
	setegid(0);
	
	if (bind(listen_sr,(struct sockaddr *)&local,sizeof(struct sockaddr))<0) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"Can't bind the minslots checking listening socket - terminating: %s.", strerror(errno));
		exit(0);
	}
		
	seteuid(conf.uid);
	setegid(conf.gid);
	
	remote_size=sizeof(struct sockaddr_in);
	/* accept connections */
	while(1)
	{
		n=recvfrom(listen_sr,bufor,10000,0,(struct sockaddr*)&remote,&remote_size);
		if (n<=0) continue;

		bufor[n]=0;

		cmd=strtok_r(bufor," ",&__strtok_temp__);
		if (strcmp(cmd,"$SR")) goto next;

		nick=strtok_r(NULL," ",&__strtok_temp__);
		if (!nick) 	continue;

		s_slots=strtok_r(NULL,"/",&__strtok_temp__);

		if (!s_slots) continue;

		s_slots=strtok_r(NULL,"\005",&__strtok_temp__);
		if (!s_slots) continue;
		if(strlen(s_slots)>3)
		{
		    pubmsg(user[n],"More than 999 slots not allowed");
		    user_set_state(user[n],STATE_QUIT_SHOW);
			continue;
		}

		ip=my_inet_ntoa(remote.sin_addr);

		pthread_mutex_lock( &mutex );

		n=nick2id(nick);

		if (n<0) goto next;
		usr=user[n];

		slots_old=usr->slots;
		usr->slots=atoi(s_slots);



		if (usr->ip && !match(usr->ip,ip)) goto next;

		if (!user_tst_state(usr,STATE_SR) || strchr(usr->perm,'s') || strchr(usr->perm,'r')) 	goto next;


		//if user was not once registered
		if(usr->slots>=conf.minslots)
		{
			if (slots_old) 
			{
				user_set_state(usr,STATE_REGISTERED);
				goto next;
			}	

			debug(DEBUG_PROTO,"RECVF %s",usr->nick);			
			user_set_state(usr,STATE_REGISTERED|STATE_SEND_NICKLIST);			
			debug(DEBUG_NET,"'%s' minslots checked",nick);
			pubmsg(usr,"The minimal amount of slots reached");
			pubmsg(usr,"-----------------------------------");

			goto next;
		}

		debug(DEBUG_PROTO,"RECVF '%s'",bufor);
		pubmsg(usr,"The minimum number of upload slots is %d.",conf.minslots);
		pubmsg(usr,"Update its number and reconnect");

		debug(DEBUG_NET,"'%s' kicked because of the amount of upload slots",nick);
		user_set_state(usr,STATE_QUIT_SHOW); 

	next:
		my_free(ip);
		pthread_mutex_unlock( &mutex );
	}
	
	debug(DEBUG_WRNG,"listen_thread_main: this should be never reached!");
	pthread_cancel(listen_th_sr); /* never reached */
} /* }}} */



/* finish() - signal handler and program termination {{{ */
void finish(int sig) 
{
	int n;
	
	switch(sig) 
	{
		/* ignore SIGPIPE */
		case SIGPIPE: return;
	
		/* rehash on SIGHUP */
		case SIGHUP: 
					debug(DEBUG_SYS|DEBUG_STD,"SIGHUP received - restarting server.");
		
					if (conf.listen_port_cons) 
					{
						pthread_cancel(listen_th_cons);
						close(listen_cons);
					}

					pthread_cancel(listen_th_main);			
					close(listen_main);

					if (conf.minslots) 
					{
						pthread_cancel(listen_th_sr);
						close(listen_sr);
					}
					seteuid(0);
					setegid(0);
					parse_config(&conf);
					make_paths(&conf);
				 
					if ((*(conf.log_main))&&(conf.log_level&DEBUG_STD) && freopen(conf.log_main,"a",stderr)==NULL) debug(DEBUG_SYS,"Couldn't open main log file (%s): %s.",conf.log_main,strerror(errno));
				 
					if (file_net)
					{
						fclose(file_net);
						file_net=NULL;
					}
				 
					if ((*(conf.log_net))&&(conf.log_level&DEBUG_NET) && (file_net=fopen(conf.log_net,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open network access log file (%s): %s.",conf.log_net,strerror(errno));
				 
					if (file_proto)
					{
						fclose(file_proto);
						file_proto=NULL;
					}
				 
					if ((*(conf.log_proto))&&(conf.log_level&DEBUG_PROTO) && (file_proto=fopen(conf.log_proto,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open protocol log file (%s): %s.",conf.log_proto,strerror(errno));
				 
					if (file_wrng)
					{
						fclose(file_wrng);
						file_wrng=NULL;
					}
				 
					if ((*(conf.log_wrng))&&(conf.log_level&DEBUG_WRNG) && (file_wrng=fopen(conf.log_wrng,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open warnings log file (%s): %s.",conf.log_wrng,strerror(errno));
				 
					if (file_con)
					{
						fclose(file_con);
						file_con=NULL;
					}
				 
					if ((*(conf.log_con))&&(conf.log_level&DEBUG_CON) && (file_con=fopen(conf.log_con,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open console log file (%s): %s.",conf.log_con,strerror(errno));
				 
					if (file_chat)
					{
						fclose(file_chat);
						file_chat=NULL;
					}
				 
					if ((*(conf.log_chat))&&(conf.log_level&DEBUG_CHAT) && (file_chat=fopen(conf.log_chat,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open chat log file (%s): %s.",conf.log_chat,strerror(errno));
				 
					seteuid(conf.uid);
					setegid(conf.gid);
				 
					if (pthread_create(&listen_th_main,NULL,(void *)listen_thread_main,NULL)!=0) 
					{
						debug(DEBUG_SYS|DEBUG_STD,"Couldn't recreate the main service thread - terminating: %s.",strerror(errno));
						exit(sig);
					}
				 
					if (conf.minslots && pthread_create( &listen_th_sr, NULL, (void *)listen_thread_SR, NULL)!=0)
					{
						debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the testing minslots thread - terminating: %s.", strerror(errno));
						finish(0);
					}

					if (conf.listen_port_cons && pthread_create(&listen_th_cons,NULL,(void *)listen_thread_cons,NULL)!=0) debug(DEBUG_SYS|DEBUG_STD,"Couldn't recreate the remote console thread - console disabled: %s.", strerror(errno));
				 
					return;
			
		case SIGSEGV: 
					debug(DEBUG_PROTO|DEBUG_SYS|DEBUG_STD|DEBUG_NET|DEBUG_CON|DEBUG_CHAT,"Crashed: Segmentation fault");	

					openlog("dcd",LOG_PID|LOG_CONS,LOG_DAEMON);
					syslog(LOG_CRIT,"Segmentation fault. Crashing!\n");
					closelog();

					exit(sig);
	
		default: 
					debug(DEBUG_SYS|DEBUG_STD,"Shutting down.");
				
					if (conf.listen_port_cons) pthread_cancel(listen_th_cons);
					pthread_cancel(listen_th_main);
			 
					for (n=0;n<NUSERS;n++) disconnect(user[n],"Hub shutdown!");
			
					exit(sig);
	}
} /* }}} */

/* buf2line() - extract single request line from buf {{{ */
char *buf2line(struct userrec *usr,char **next) 
{
	char *line;
	size_t holdlen=0;

	if (usr->bufrd) 
	{ 
		my_free(usr->buf); 
		usr->buf=NULL; 
		usr->bufrd=0; 
	}
	
	if (*next==NULL) return NULL;
	if (**next==0) return NULL;

	line=*next;
	
	if (usr->cons) *next=strpbrk(line,"\r\n"); 
	else *next=strchr(line,'|');
	
	if ((usr->buf)||(*next==NULL)) 
	{
		if (*next) **next=0;
		if (usr->buf) holdlen=strlen(usr->buf);
	
		usr->buf=my_realloc(usr->buf,holdlen+strlen(line)+1);
	
		if (usr->buf==NULL) 
		{ 
			if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) user_set_state(usr,STATE_QUIT_SHOW);
			else  user_set_state(usr,STATE_QUIT);

			pubmsg(usr,"Kicked - excess flood");
			return NULL; 
		}
	
		if (holdlen==0) *(usr->buf)=0;
		
		strcat(usr->buf,line);
		if (*next==NULL) return NULL;
	}
	
	if (usr->cons)
	{ 
		while ((**next=='\r')||(**next=='\n')) 
		{ 
			**next=0; 
			(*next)++; 
		}
	}
	else 
		{ 
			**next=0; 
			(*next)++; 
		}

	if (usr->buf) 
	{ 
		usr->bufrd=1; 
		line=usr->buf; 
	}
	
	return line;
} /* }}} */

//used for testing if user has more than minslots open

/* manage_sock() - proces requests remaining on a socket {{{ */
void manage_sock(struct userrec *usr) 
{
	const char *commands[]=
	{
		"Key",
		"Quit",
		"ValidateNick",
		"GetNickList",
		"MyINFO",
		"GetINFO",
		"Kick",
		"MyPass",
		"Search",
		"MultiSearch",
		"To:",
		"Version",
		"ConnectToMe",
		"MultiConnectToMe",
		"RevConnectToMe",
		"UniSearch",
		"SR",
		"OpForceMove",
		"Supports",
		"" /* indicates end of list */
	};


	const char *sep=" ";
	char *buf=NULL, *line=NULL, *next=NULL, *line_copy=NULL, *cmd=NULL, *par=NULL;
	int pos=0,count,n;
	char *tmp1=NULL, *tmp2=NULL,*tmp3=NULL,*__strtok_temp__;

	debug(DEBUG_PROTO,"manage_sock() enter");

	/* read all available data */
	n=128;
	buf=(char *)my_malloc(n);
	
	while(1) 
	{
		count=recv(usr->sock,buf+pos,n-pos,0); //get the text
		if (count>0) pos+=count;
		if (((count>0)&&(pos<n))||((count<0)&&(errno==EAGAIN))) break;
	
		if (count<0 && errno!=EAGAIN) 
		{ 
#if defined(WIN32) || defined(__CYGWIN__)
			if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) user_set_state(usr,STATE_QUIT_SHOW);
			else user_set_state(usr,STATE_QUIT);
	
			goto leave; 
#else
			debug(DEBUG_WRNG,"recv() failed from %s@%s: %s.",(usr->nick)?usr->nick:"Unknown",usr->ip,strerror(errno)); 
			break; 
#endif
		}

		if (count==0) 
		{ 
			if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) user_set_state(usr,STATE_QUIT_SHOW);
			else  user_set_state(usr,STATE_QUIT);
	
			goto leave; 
		}

		n*=2;
		buf=my_realloc(buf,n);
		if (buf==NULL) 
		{ 
			if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) user_set_state(usr,STATE_QUIT_SHOW);
			else  user_set_state(usr,STATE_QUIT);
		
			pubmsg(usr,"Kicked - excess flood");

			goto leave; 
		}
	}

	if (pos<=0) goto leave;
	buf=my_realloc(buf,pos+1);
	buf[pos]=0;

	/* update last ping */
	if (user_tst_state(usr,STATE_REGISTERED)) usr->ping=time(NULL);

	/* parse each line in loop */
	next=buf;
	
	while((line=(char *)buf2line(usr,&next))) //makes a line from buf
	{
		my_free(line_copy);
		line_copy=strdup(line);

		if (usr->cons) 
		{
			if (manage_cons(usr,line_copy)) goto leave;  //
			else continue;
		}

		debug(DEBUG_PROTO,"RECV '%s'",line);		

		if (!(cmd=strtok_r(line_copy,sep,&__strtok_temp__))) continue;
		if (!*cmd) continue;
		par=strtok_r(NULL,sep,&__strtok_temp__);
		count=0; /* - $ command number*/
	
		/* Parse commands {{{ */
		if (*cmd=='$')
		{
			while(*commands[count]) 
			{
				if (!strcasecmp(&cmd[1],commands[count])) break;
				else count++; 
			}
		}
		else
			if ((*cmd=='<')&&(cmd[strlen(cmd)-1]=='>')) count=-1;
			else count=-2; // unrecognized command


		if ((user_tst_state(usr,STATE_CONNECTED) && count!=0 && count!=18) || 
			(user_tst_state(usr,STATE_KEY) && count!=2 )) 
		{
			user_set_state(usr,STATE_QUIT);
			pubmsg(usr,"Broken registration.");
			goto leave; 
		}

		switch(count) 
		{
			/* -1. pubmsg {{{ */
			case -1: 
					if (!user_tst_state(usr,STATE_REGISTERED)) continue;

					cmd[strlen(cmd)-1]=0; 
					cmd++;
			
					if (!strcasecmp(cmd,usr->nick)) command(usr,line);
					else debug(DEBUG_PROTO,"On PublicMsg from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip, line);	

					continue; /* }}} */
						 
			/* 0. $Key {{{ */
			case 0: 
					if (!user_tst_state(usr,STATE_CONNECTED)) continue;
					if (!validkey(par)) 
					{
						pubmsg(usr,"Invalid $Key reply.");
						user_set_state(usr,STATE_QUIT);
						debug(DEBUG_PROTO,"On Key from %s@%s: Invalid reply. (%s)",usr->nick,usr->ip,line);	
						goto leave;
					}
			
					user_set_state(usr,STATE_KEY);
					continue; /* }}} */
						
			/* 1. $Quit {{{ */
			case 1: 
					if (!par) continue;
					if (strcasecmp(par,usr->nick)) 
					{
						debug(DEBUG_PROTO,"On quit from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
							continue;
					}
				
					if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) user_set_state(usr,STATE_QUIT_SHOW);
					else user_set_state(usr,STATE_QUIT);	

					goto leave; /* }}} */
					
			/* 2. $ValidateNick {{{ */
			case 2: 
					if (!par) continue;
		
					if (user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) 
					{
						pubmsg(usr,"Your nick has already been validated.");
						debug(DEBUG_PROTO,"On ValidateNick from %s@%s: Nick has already been validated. (%s)",usr->nick,usr->ip,line);
						continue;
					}
				
					debug(DEBUG_NET,"'%s' from %s is validating",par,usr->ip);	

					switch (validnick(usr,par)) 
					{
						case 0: 
								user_set_state(usr,STATE_QUIT);
								debug(DEBUG_NET,"'%s' not loged in",usr->nick);
								goto leave;
					
						case 1:	
								/* get the penalties for the user*/
								if (penalty_welcome_banned(usr)) continue;

								my_free(usr->nick);
								usr->nick=strdup(par);
								user_set_state(usr,STATE_LOGGEDIN);

								debug(DEBUG_NET,"'%s' logged in",usr->nick);
								sendf(usr,"$Hello %s|",usr->nick);
								continue;	

						case 2:
								my_free(usr->nick);
								usr->nick=strdup(par);

								if (penalty_welcome_banned(usr)) continue;
								
								user_set_state(usr,STATE_PASSWORD);
								debug(DEBUG_NET,"'%s' requires password",usr->nick);
								sendf(usr,"$GetPass|");
								continue;
					}   
			/* }}} */
					
			/* 3. $GetNickList {{{ */
			case 3: 
					if (user_tst_state(usr,STATE_LOGGEDIN) && (conf.flags&INFO_BOOST)) continue;
				
					if (conf.minslots && 
						!user_tst_state(usr,STATE_REGISTERED) && 
						!strchr(usr->perm,'s')  && 
						!strchr(usr->perm,'r')) 
					{
						sendf(usr,"$NickList %s$$|",usr->nick);
						continue;
					}

					GetNickList(usr);
					continue; /* }}} */
					
			/* 4. $MyINFO {{{ */
			case 4: 
					if(!par || !user_tst_state(usr,STATE_LOGGEDIN|STATE_REGISTERED)) continue;
						
					if (fill_info(usr,line)) 
					{
						if (!strchr(usr->perm,'s') && !strchr(usr->perm,'m') && usr->share<(conf.minshare*1024*1024)) 
						{
							sendf(usr,"<Hub> You must share at least %lluMB on this hub.|",conf.minshare);
							user_set_state(usr,STATE_QUIT_SHOW);
							goto leave;
						}
													
//						sendf(NULL,"%s|",line);
					
						if (user_tst_state(usr,STATE_LOGGEDIN)) 
						{
							if (*(conf.conf_motd)) playfile(usr,conf.conf_motd);
							if (conf.flags & SHOW_JQ) sendf(NULL,"## %s has joined|",usr->nick);
												
							if (conf.flags & INFO_BOOST)
								for(n=0;n<NUSERS;n++)
									if (user[n] && user_tst_state(user[n],STATE_REGISTERED))
										sendf(usr,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",user[n]->nick,user[n]->desc,user[n]->ctype,user[n]->flag,user[n]->email,user[n]->share);
						}
						
//						sendf(NULL,"%s|",line);

						if (conf.minslots && !strchr(usr->perm,'r') && !strchr(usr->perm,'s'))
						{
							if (!usr->slots)	
							{	
								pubmsg(usr,"===============================");
								pubmsg(usr,"Checking for slots. Please wait.");
								pubmsg(usr,"===============================");
							}

							user_set_state(usr,STATE_SR|usr->state);

							//wrapper for dctc and dcgui-qt client
							sendf(usr,"$Search %s:%d F?F?0?1?fakesearch$$|",usr->con_ip,conf.listen_port_sr);

							//dc++, dc
							sendf(usr,"$Search %s:%d F?F?0?1?.$$|",usr->con_ip,conf.listen_port_sr);
							sendf(usr,"%s|",line);
							usr->ping=time(NULL);
							goto leave;	
						}

						user_set_state(usr,STATE_REGISTERED);						

						for(n=0;n<NUSERS;n++)
							if (user[n] && user_tst_state(user[n],STATE_REGISTERED))
								sendf(usr,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",user[n]->nick,user[n]->desc,user[n]->ctype,user[n]->flag,user[n]->email,user[n]->share);

						sendf(NULL,"%s|",line);
						goto leave;
					}
					
					user_set_state(usr,STATE_QUIT_SHOW);
					pubmsg(usr,"Invalid MyINFO received.");
					goto leave; 
						
					continue; /* }}} */
					
			/* 5. $GetINFO {{{ */
			case 5: 
					if (conf.flags&INFO_BOOST) continue;
					
					if(!par || !user_tst_state(usr,STATE_LOGGEDIN|STATE_SR|STATE_REGISTERED)) continue;

					n=nick2id(par);
					if (n<0)	
					{
						debug(DEBUG_PROTO,"On GetINFO from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						continue;
					}

					if (user[n] && user_tst_state(user[n],STATE_REGISTERED))
						sendf(usr,"$MyINFO $ALL %s %s$ $%s%c$%s$%llu$|",user[n]->nick,user[n]->desc,user[n]->ctype,user[n]->flag,user[n]->email,user[n]->share);
					continue; /* }}} */
					
			/* 6. $Kick {{{ */ 
			case 6:
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) continue;
	
					if(!strchr(usr->perm,'a') && !strchr(usr->perm,'s'))
					{
						debug(DEBUG_NET,"%s(not OP) tried to kick (%s)",usr->nick,par);
						pubmsg(usr,"You are Not an Op");
						continue;
					}

					if (!strcmp(usr->nick,par))
					{
						privmsg(usr,NULL,"You are trying to kick yourself");
						continue;
					}


					if ((n=nick2id(par))<0)
					{
						pubmsg(usr,"You are trying to kick a person, that doesn't exist");
						continue;
					}
	
					if (usr->perm!=NULL && user[n]->perm!=NULL)
					{
						if (strchr(user[n]->perm,'a') && strchr(usr->perm,'a')) 
						{  
							privmsg(usr,NULL,"You are trying to kick OP");
							continue;
						}
			
						if (strchr(user[n]->perm,'s') && (strchr(usr->perm,'a') || strchr(usr->perm,'s'))) 
						{
							privmsg(usr,NULL,"You are trying to kick sOP");
							continue;
						}
					
						if (strchr(user[n]->perm,'a') && strchr(usr->perm,'s')) privmsg(usr,NULL,"You are kicking OP");
					}
					else continue;


					debug(DEBUG_NET,"'%s'(OP) kicked '%s'",usr->nick,par);
					privmsg(user[n],NULL,"You were Kicked");

	
					user_set_state(user[n],STATE_QUIT_SHOW);
					continue; /* }}} */
						
			/* 7. $MyPass {{{ */
			case 7:
					if(!par || !user_tst_state(usr,STATE_PASSWORD)) continue;

					if (!strcmp(usr->password,par)) 
					{
						debug(DEBUG_NET,"'%s' sent correct password", usr->nick);
						pubmsg(usr,"Your password is correct");
				
						if (usr->perm!=NULL)
						{	
							if (strchr(usr->perm,'a'))
							{  
								pubmsg(usr,"You become an OP");
								debug(DEBUG_NET,"%s becomes an OP", usr->nick);
							}
	
							if (strchr(usr->perm,'s'))
							{  
								pubmsg(usr,"You become an sOP");
								debug(DEBUG_NET,"%s becomes an sOP", usr->nick);
							}
								
							sendf(usr,"$LogedIn|");
						}
							
						sendf(usr,"$Hello %s|",usr->nick);
						debug(DEBUG_NET,"'%s' is in", usr->nick);
		
						user_set_state(usr,STATE_LOGGEDIN);

					}else 
						{	
							debug(DEBUG_NET,"'%s' sent incorrect password", usr->nick);
							pubmsg(usr,"Your password is incorrect");
							sendf(usr,"$BadPass|$GetPass|");
							continue;
						}
						
					continue; /* }}} */
					
			/* 8. $Search {{{ */
			case 8: 
					if(!par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NOSEARCH)) continue;
	
					if (!(cmd=strtok_r(par,":",&__strtok_temp__))) continue;
			
					if (!strcasecmp(cmd,"Hub")) 
					{
					
						//if passive mode is not allowed and user has no permission to be passive
						if (!conf.allow_passive && !strchr(usr->perm,'p'))
						{
							privmsg(usr,NULL,"PASSIVE mode is not allowed.");
							privmsg(usr,NULL,"Change your connection settings and reconnect.");
						
							//kicking user from hub
							user_set_state(usr,STATE_QUIT_SHOW);
							continue;
						} 
					
						if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) continue;
	
						if (strcasecmp(usr->nick,par)) 
						{
							debug(DEBUG_PROTO,"On Search from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
							continue;
						}
			
					}else 
						{
							if (strcmp(usr->ip,cmd)) 
							{
								debug(DEBUG_PROTO,"On Search from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
								continue;
							}
						}

					sendf(NULL,"%s|",line);
				
					continue; /* }}} */

			/* 9. $MultiSearch {{{ */
			case 9:
					debug(DEBUG_PROTO,"On MultiSearch form %s@%s: Not supported yet. (%s)",usr->nick,usr->ip,line);
					continue; /* }}} */

			/* 10. $To: {{{ */
			case 10: 
					if ((n=nick2id(par))<0)
					{
						/* get rid of some client-side junk */
						if (strcmp(par,"Vandel\\Debug")) debug(DEBUG_PROTO,"On PrivMsg from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) continue;
					if (strcasecmp(par,"From:")) continue;
					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) continue;

					if (strcasecmp(usr->nick,par)) 
					{
						debug(DEBUG_PROTO,"On PrivMsg from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					if (!(par=strtok_r(NULL," $",&__strtok_temp__))) continue;
					if (!(par=strtok_r(NULL,"",&__strtok_temp__))) continue;
			
			
					// if the user[n] is not OP or sOP, and usr has no permission to private chat then ontinue;
					if (user_tst_penalty(usr,PENALTY_NOPRVCHAT) && !strchr(user[n]->perm,'s') && !strchr(user[n]->perm,'s')) continue;

					if (user[n]->cons) sendf(user[n],"[%s@%s] %s\r\n",usr->nick,usr->ip,par);
					else sendf(user[n],"%s|",line);
				
					continue; /* }}} */
					 
			/* 11. $Version {{{ */
			case 11: 
					if (!par) continue;
					if (!*par) continue;
					if (strlen(par)+1>VER_LEN) continue;
					usr->ver=strdup(par);
					continue; /* }}} */
					 
			/* 12. $ConnectToMe {{{ */
			case 12:
					if(!par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NODL)) continue;
				
					if ((n=nick2id(par))<0) 
					{
						debug(DEBUG_PROTO,"On ConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					if (user[n]->cons)
					{ 
						privmsg(usr,NULL,"Can't connect to console user.");
						continue; 
					}
				
					if (!(par=strtok_r(NULL,":",&__strtok_temp__))) continue;
					if (strcmp(usr->ip,par)) 
					{
						debug(DEBUG_PROTO,"On ConnectToMe from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					sendf(user[n],"%s|",line);
					continue; /* }}} */
					 
			/* 13. $MultiConnectToMe {{{ */
			case 13:
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) continue;
	
					if ((n=nick2id(par))<0)
					{
						debug(DEBUG_PROTO,"On MultiConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					if (user[n]->cons)
					{
						privmsg(usr,NULL,"Can't connect to console user.");
						continue; 
					}
			
					if (!(par=strtok_r(NULL,":",&__strtok_temp__))) continue;
				
					if (strcmp(usr->ip,par)) 
					{
						debug(DEBUG_PROTO,"On MultiConnectToMe from %s@%s: Incorrect IP. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					sendf(user[n],"%s|",line);
					continue; /* }}} */
					 
			/* 14. $RevConnectToMe {{{ */
			case 14: 
					if(!par || !user_tst_state(usr,STATE_REGISTERED) || user_tst_penalty(usr,PENALTY_NODL)) continue;
	
					//if passive mode is not allowed and user has no permission to be passive
					if (!conf.allow_passive && !strchr(usr->perm,'p'))
					{
						privmsg(usr,NULL,"PASSIVE mode is not allowed.");
						privmsg(usr,NULL,"Change your connection settings and reconnect.");
						
						//kicking user from hub
						user_set_state(usr,STATE_QUIT_SHOW);
						continue;
					} 
			
					if (strcasecmp(usr->nick,par)) 
					{
						debug(DEBUG_PROTO,"On RevConnectToMe from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						continue;
					}
				
					if (!(par=strtok_r(NULL,sep,&__strtok_temp__))) continue;
			
					if ((n=nick2id(par))<0) 
					{
						debug(DEBUG_PROTO,"On RevConnectToMe from %s@%s: No such user. (%s)",usr->nick,usr->ip,line);
						continue;
					}
			
					if (user[n]->cons)
					{
						privmsg(usr,NULL,"Can't connect to console user.");
						continue; 
					}
			
					sendf(user[n],"%s|",line);
					continue; /* }}} */
					 
			/* 15. $UniSearch {{{ */
			case 15: 
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) continue;
	
					debug(DEBUG_PROTO,"On UniSearch from %s@%s: Not supported yet. (%s)",usr->nick,usr->ip,line);
					privmsg(usr,NULL,"UniSearch not supported yet.");
					continue; /* }}} */
					 
			/* 16. $SR {{{ */
			case 16: 
					if(!par || !user_tst_state(usr,STATE_REGISTERED)|| user_tst_penalty(usr,PENALTY_NOSEARCH) ) continue;
	
					if (strcasecmp(usr->nick,par)) 
					{
						debug(DEBUG_PROTO,"On SR from %s@%s: Fake nick not allowed. (%s)",usr->nick,usr->ip,line);
						continue;
					}
						 
					my_free(line_copy);
					line_copy=strdup(line);
					 
					par=strrchr(line_copy,'\005');
					if(!par) continue;

					*par=0;
					par++;
					n=nick2id(par);
					if (n<0) continue;
					
					sendf(user[n],"%s|",line_copy);
					continue; /* }}} */
					
			/* 17. $OpForceMove {{{ */
			case 17: 
					if(!par || !user_tst_state(usr,STATE_REGISTERED)) continue;
	
					if (!strchr(usr->perm,'a') && !strchr(usr->perm,'s'))
					{  
						debug(DEBUG_NET,"'%s'(not OP) tried to redirect someone",usr->nick);
						pubmsg(usr,"You are Not an Op");
						continue;
					}
						 						 
					my_free(line_copy);
					line_copy=strdup(line);
						
					tmp1=strtok_r(line_copy," ",&__strtok_temp__); //$OpForceMove:
					tmp1=strtok_r(NULL,":",&__strtok_temp__); //$Who:
					tmp1=strtok_r(NULL,"$",&__strtok_temp__);//<victimNick>
	
					tmp2=strtok_r(NULL,":",&__strtok_temp__);//$Where:
					tmp2=strtok_r(NULL,"$",&__strtok_temp__);// <IP>
					
					tmp3=strtok_r(NULL,":",&__strtok_temp__)+4;// $Msg:<message>
																
					n=nick2id(tmp1);
					 
					if (!strcmp(usr->nick,tmp1))
					{
						privmsg(usr,NULL,"You are trying to move yourself");
						continue;
					}

					if (n<0)
					{
						pubmsg(usr,"You are trying to redirect person, that doesn't exist");
						continue;
					}					

					if (usr->perm!=NULL && user[n]->perm!=NULL)
					{
						if (strchr(user[n]->perm,'a') && strchr(usr->perm,'a')) 
						{  
							privmsg(usr,NULL,"You are trying to redirect OP");
							continue;
						}
						
						if (strchr(user[n]->perm,'s') && (strchr(usr->perm,'a') || strchr(usr->perm,'s'))) 
						{
							privmsg(usr,NULL,"You are trying to redirect sOP");
							continue;
						}
	
						if (strchr(user[n]->perm,'a') && strchr(usr->perm,'s')) privmsg(usr,NULL,"You are redirecting OP");
					}
					else continue;

					privmsg(user[n],NULL,tmp3);
					sendf(user[n],"$ForceMove %s|",tmp2);			
					if((n=nick2id(tmp1))>=0) user_set_state(user[n],STATE_QUIT_SHOW);
					
					debug(DEBUG_NET,"'%s'(OP) moved '%s' because: '%s'",usr->nick,tmp1,tmp3);
					continue; /* }}} */

			case 18:
					sendf(usr,"$Supports NoGetINFO NoHELLO|");
					continue;

			default: 
					debug(DEBUG_PROTO,"On %s from %s@%s: Unknown command. (%s)",cmd,usr->nick,usr->ip,line);
		}
		/* }}} */
	}
	
leave:
//		my_free(line);
		my_free(line_copy);
		my_free(buf);
		
		debug(DEBUG_PROTO,"manage_sock() leave");
} /* }}} */

/* init_var() - install the signal handler and initialize variables {{{ */
void init_var(char *argv0) 
{
	struct passwd *user_data;
	struct group *group_data;

	/* install signal handler */
	signal(SIGPIPE,finish);
	signal(SIGHUP,finish);
	signal(SIGINT,finish);
	signal(SIGTERM,finish);
	signal(SIGSEGV,finish);
	
	/* clear main variables */
	FD_ZERO(&sockfdset);
	NUSERS=0;
	NPENAL=0;
	DAEMON=0;
	
	bzero(penalties,MAXPENALTIES*sizeof(Tpenalty *));
	bzero(user,MAXUSERS*sizeof(struct userrec *));
	bzero(&conf,sizeof(struct config));


	/* set up default config values */

	conf.conf_main=(char *)my_malloc(strlen(CONFIGDIR)+strlen(CONF_MAIN)+2); /* one byte for '/' and one for '\0' */
	sprintf(conf.conf_main, "%s%s%s", CONFIGDIR, (CONFIGDIR[strlen(CONFIGDIR)-1]=='/')?"":"/", CONF_MAIN);

	conf.conf_welcome=strdup(CONF_WELCOME);
	conf.conf_motd=strdup(CONF_MOTD);
	conf.conf_banned=strdup(CONF_BANNED);
	conf.conf_cusers=strdup(CONF_CUSERS);
	conf.conf_callow=strdup(CONF_CALLOW);
	conf.conf_nallow=strdup(CONF_NALLOW);
	conf.conf_penalties=strdup(CONF_PENALTIES);
	conf.debug_level=DEBUG_LEVEL;
	conf.log_level=LOG_LEVEL;
	conf.log_dir=strdup(LOG_DIR);
	conf.log_main=strdup(LOG_MAIN);
	conf.log_net=strdup(LOG_NET);
	conf.log_proto=strdup(LOG_PROTO);
	conf.log_wrng=strdup(LOG_WRNG);
	conf.log_con=strdup(LOG_CON);
	conf.log_chat=strdup(LOG_CHAT);
	conf.listen_interface=strdup(LISTEN_INTERFACE);
	conf.admin_contact=strdup(ADMIN_CONTACT);
	conf.listen_port_main=LISTEN_PORT_MAIN;
	conf.listen_port_cons=LISTEN_PORT_CONS;

	user_data=getpwnam(DEFAULT_USER);
	if(!user_data) 
	{
		debug(DEBUG_STD,"User `%s' not found - terminating.",DEFAULT_USER);
		exit(0);
	}
	
	conf.uid=user_data->pw_uid;

	group_data=getgrnam(DEFAULT_GROUP);
	
	if(!group_data) 
	{
		debug(DEBUG_STD,"Group `%s' not found - terminating.",DEFAULT_GROUP);
		exit(0);
	}
	
	conf.gid=group_data->gr_gid;

	conf.hubname=strdup(HUBNAME);
	conf.userlimit=USERLIMIT;
	conf.allow_broken_key=ALLOW_BROKEN_KEY;
	conf.allow_chat=ALLOW_CHAT;
	conf.allow_passive=ALLOW_PASSIVE;
	conf.std_penalty_duration=STD_PENALTY_DURATION;

	conf.ping_timeout=PING_TIMEOUT;
	conf.register_timeout=REGISTER_TIMEOUT;

	/* end of default config values */
	
	/* logging file handlers */
	file_net=NULL;
	file_proto=NULL;
	file_wrng=NULL;
	file_chat=NULL;

	/* tcpd syslog messages */
	allow_severity=LOG_INFO;
	deny_severity=LOG_WARNING;

	/* other */
	__progname=get_progname(argv0);
	randomize();
	time(&starttime);
} /* }}} */

/* main() - main program {{{ */
int main(int argc,char *argv[]) 
{
	int i,maxfd;
	struct timeval timeout;
	fd_set readfdset;

	/* init variables */
	init_var(argv[0]);
	maxfd=getdtablesize();
	if(maxfd>FD_SETSIZE)
	maxfd=FD_SETSIZE;

	/* parse command-line options */
	parse_options(&conf,argc,argv);

	/* read config file */
	if (parse_config(&conf)) exit(0);

	/* compose direct paths to files for further use */
	make_paths(&conf);


	/* test if we're allowed to bind to requested ports here,
	because we can't see stderr after daemoN() */
	bind_test(conf.listen_port_main);
	if (conf.listen_port_cons) bind_test(conf.listen_port_cons);


	/* set effective uid/gid to proper user/group */
	seteuid(conf.uid);
	setegid(conf.gid);

	if (conf.userlimit>MAXUSERS) 
	{
	debug(DEBUG_STD,"User limit (%d) is higher than MAXUSERS (%d). See main.h and recompile.",conf.userlimit,MAXUSERS);
	exit(0);
	}

	/* daemonize */
	if (daemon(0,0)<0) debug(DEBUG_STD,"Couldn't fall into background: %s.", strerror(errno));
	else 
		{
			DAEMON=1;
			seteuid(0);
			setegid(0);
		
			if ((*(conf.log_main))&&(conf.log_level&DEBUG_STD))
			if (freopen(conf.log_main,"a",stderr)==NULL)
				debug(DEBUG_SYS,"Couldn't open main log file (%s): %s.", conf.log_main, strerror(errno));
		
			seteuid(conf.uid);
			setegid(conf.gid);
			debug(DEBUG_SYS|DEBUG_STD,"Daemon started.");
		}

		/* open rest of the log files */
	seteuid(0);
	setegid(0);
	
	if ((*(conf.log_net))  && (conf.log_level&DEBUG_NET) && (file_net=fopen(conf.log_net,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open network access log file (%s): %s.",conf.log_net,strerror(errno));
	
	if ((*(conf.log_proto)) && (conf.log_level&DEBUG_PROTO) && (file_proto=fopen(conf.log_proto,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open protocol log file (%s): %s.",conf.log_proto,strerror(errno));
	
	if ((*(conf.log_wrng)) && (conf.log_level&DEBUG_WRNG) && (file_wrng=fopen(conf.log_wrng,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open warnings log file (%s): %s.",conf.log_wrng,strerror(errno));
	
	if ((*(conf.log_con)) && (conf.log_level&DEBUG_CON) && (file_con=fopen(conf.log_con,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open console log file (%s): %s.",conf.log_con,strerror(errno));
	
	if ((*(conf.log_chat)) && (conf.log_level&DEBUG_CHAT) && (file_chat=fopen(conf.log_chat,"a"))==NULL) debug(DEBUG_SYS|DEBUG_STD,"Couldn't open chat log file (%s): %s.",conf.log_chat,strerror(errno));
	
	seteuid(conf.uid);
	setegid(conf.gid);

	/* initialize semaphores */
	if (pthread_mutex_init(&mutex, 0) ||
		pthread_mutex_init(&mutex_inet_ntoa, 0))
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create synchronization mutex: %s - terminating.", strerror(errno)); 
		finish(0);
	}

	/* create listening thread for main service */
	if (pthread_create( &listen_th_main, NULL, (void *)listen_thread_main, NULL)!=0) 
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the main service thread - terminating: %s.", strerror(errno));
		finish(0);
	}


	if (conf.minslots && pthread_create( &listen_th_sr, NULL, (void *)listen_thread_SR, NULL)!=0)
	{
		debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the testing minslots thread - terminating: %s.", strerror(errno));
		finish(0);
	}

	/* create listening thread for remote console */
	if (conf.listen_port_cons && pthread_create( &listen_th_cons, NULL, (void *)listen_thread_cons, NULL)!=0) debug(DEBUG_SYS|DEBUG_STD,"Couldn't create the remote console thread - console disabled: %s.", strerror(errno));

	/* reading the penalties file */
	penalties_read();

	/* main loop - finish only when error or killed with a signal */
	while(1) 
	{
	/* we use 1ms timeout because we want select() to proces new users even when
	 * there is no traffic on the sockets */

		timeout.tv_sec=0; timeout.tv_usec=1000;
		bcopy(&sockfdset,&readfdset,sizeof(readfdset));

		if (select(maxfd,&readfdset,NULL,NULL,&timeout)<0) 
		{
			if (errno!=EINTR) debug(DEBUG_WRNG,"select() failed (%s).", strerror(errno));
			continue;
		}

		/* this needs to be a reverse loop because of the method used in deluser() */
		for(i=NUSERS-1;i>=0;i--) 
		{
			if (user[i])
			{
				if(FD_ISSET(user[i]->sock,&readfdset)) manage_sock(user[i]);
				check_timeout(user[i]);
				if(user_tst_state(user[i],STATE_SEND_NICKLIST)) GetNickList(user[i]);
				if(user_tst_state(user[i],STATE_QUIT)) deluser(user[i],NULL);
			}
		}
	}

	debug(DEBUG_WRNG,"main(): this should be never reached!");
	finish(0); /* never reached */
	return 0;
} /* }}} */

/* VIM Settings {{{
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * soft-stop-width: 4
 * c indent on
 * End:
 * vim600: sw=4 ts=4 sts=4 cindent fdm=marker
 * vim<600: sw=4 ts=4
 * }}} */
